"""
Core functionality for the FFmpeg backend.

Contains utility functions for finding ffmpeg/ffprobe, running commands, etc.
"""

import os
import sys
import subprocess
import tempfile
import shutil
import json
import logging
from typing import List, Optional, Callable, Dict, Any, Union
from pathlib import Path

logger = logging.getLogger(__name__)

class FFmpegCore:
    """Core FFmpeg functionality."""

    def _find_ffmpeg(self) -> str:
        """
        Find the FFmpeg executable.

        Returns:
            Path to FFmpeg executable

        Raises:
            ValueError: If FFmpeg is not found
        """
        # Check environment variable first
        env_path = os.environ.get("FMUSVID_FFMPEG_PATH")
        if env_path and os.path.isfile(env_path):
            logger.info(f"Found FFmpeg from environment: {env_path}")
            return env_path

        # Try to find ffmpeg in PATH
        ffmpeg_path = shutil.which("ffmpeg")

        if ffmpeg_path:
            logger.info(f"Found FFmpeg at: {ffmpeg_path}")
            return ffmpeg_path

        # Try common installation locations
        common_locations = []

        if sys.platform == "win32":
            common_locations = [
                r"C:\Program Files\FFmpeg\bin\ffmpeg.exe",
                r"C:\Program Files (x86)\FFmpeg\bin\ffmpeg.exe",
                r"C:\Program Files\ImageMagick-7.1.0-Q16-HDRI\ffmpeg.exe",
                os.path.expanduser(r"~\FFmpeg\bin\ffmpeg.exe")
            ]
        elif sys.platform == "darwin":  # macOS
            common_locations = [
                "/usr/local/bin/ffmpeg",
                "/opt/homebrew/bin/ffmpeg",
                "/opt/local/bin/ffmpeg"
            ]
        else:  # Linux and others
            common_locations = [
                "/usr/bin/ffmpeg",
                "/usr/local/bin/ffmpeg",
                "/opt/ffmpeg/bin/ffmpeg"
            ]

        for location in common_locations:
            if os.path.isfile(location):
                logger.info(f"Found FFmpeg at: {location}")
                return location

        # If we get here, FFmpeg was not found
        raise ValueError(
            "FFmpeg not found. Please install FFmpeg and make sure it's in your PATH, "
            "or specify the path to FFmpeg using FMUSVID_FFMPEG_PATH environment variable."
        )

    def _find_ffprobe(self) -> str:
        """
        Find the FFprobe executable.

        Returns:
            Path to FFprobe executable

        Raises:
            ValueError: If FFprobe is not found
        """
        # Check environment variable first
        env_path = os.environ.get("FMUSVID_FFPROBE_PATH")
        if env_path and os.path.isfile(env_path):
            logger.info(f"Found FFprobe from environment: {env_path}")
            return env_path

        # Try to find ffprobe in PATH
        ffprobe_path = shutil.which("ffprobe")

        if ffprobe_path:
            logger.info(f"Found FFprobe at: {ffprobe_path}")
            return ffprobe_path

        # Try to find ffprobe in the same directory as ffmpeg
        if self._ffmpeg_path:
            ffmpeg_dir = os.path.dirname(self._ffmpeg_path)
            ffprobe_candidate = os.path.join(ffmpeg_dir, "ffprobe")

            if sys.platform == "win32":
                ffprobe_candidate += ".exe"

            if os.path.isfile(ffprobe_candidate):
                logger.info(f"Found FFprobe at: {ffprobe_candidate}")
                return ffprobe_candidate

        # Try common installation locations
        common_locations = []

        if sys.platform == "win32":
            common_locations = [
                r"C:\Program Files\FFmpeg\bin\ffprobe.exe",
                r"C:\Program Files (x86)\FFmpeg\bin\ffprobe.exe",
                r"C:\Program Files\ImageMagick-7.1.0-Q16-HDRI\ffprobe.exe",
                os.path.expanduser(r"~\FFmpeg\bin\ffprobe.exe")
            ]
        elif sys.platform == "darwin":  # macOS
            common_locations = [
                "/usr/local/bin/ffprobe",
                "/opt/homebrew/bin/ffprobe",
                "/opt/local/bin/ffprobe"
            ]
        else:  # Linux and others
            common_locations = [
                "/usr/bin/ffprobe",
                "/usr/local/bin/ffprobe",
                "/opt/ffmpeg/bin/ffprobe"
            ]

        for location in common_locations:
            if os.path.isfile(location):
                logger.info(f"Found FFprobe at: {location}")
                return location

        # If we get here, FFprobe was not found
        # Try to see if we can use ffmpeg -version to get some info instead
        logger.warning("FFprobe not found. Some features may not work properly.")

        # Return ffmpeg path as fallback (for basic functionality)
        if self._ffmpeg_path:
            return self._ffmpeg_path

        raise ValueError(
            "FFprobe not found. Please install FFmpeg (which includes FFprobe) "
            "and make sure it's in your PATH, or specify the path to FFprobe using "
            "FMUSVID_FFPROBE_PATH environment variable."
        )

    def _ensure_temp_dir(self):
        """Ensure that a temporary directory exists."""
        if self._temp_dir is None or not os.path.isdir(self._temp_dir):
            self._temp_dir = tempfile.mkdtemp(prefix="fmusvid_")
            logger.debug(f"Created temporary directory: {self._temp_dir}")

    def _run_ffmpeg(self, args: List[str], input_data=None,
                   progress_callback: Optional[Callable[[float], None]] = None) -> bytes:
        """
        Run FFmpeg with the given arguments.

        Args:
            args: FFmpeg command-line arguments
            input_data: Input data to pipe to FFmpeg
            progress_callback: Function to call with progress (0-1)

        Returns:
            FFmpeg output

        Raises:
            RuntimeError: If FFmpeg returns a non-zero exit code
        """
        # Add FFmpeg executable to the beginning of the arguments
        cmd = [self._ffmpeg_path] + args

        # Add progress monitoring if requested
        if progress_callback:
            cmd += ["-progress", "pipe:1"]

        # Log the command
        logger.debug(f"Running FFmpeg: {' '.join(cmd)}")

        # Set up the subprocess
        stdin = subprocess.PIPE if input_data is not None else None

        # Run FFmpeg
        process = subprocess.Popen(
            cmd,
            stdin=stdin,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=False
        )

        # Send input data if provided
        if input_data is not None:
            stdout, stderr = process.communicate(input=input_data)
        else:
            # Monitor progress if callback provided
            if progress_callback:
                stdout, stderr = self._monitor_progress(process, progress_callback)
            else:
                stdout, stderr = process.communicate()

        # Check return code
        if process.returncode != 0:
            error_msg = stderr.decode("utf-8", errors="replace")
            logger.error(f"FFmpeg error (code {process.returncode}): {error_msg}")
            raise RuntimeError(f"FFmpeg error (code {process.returncode}): {error_msg}")

        return stdout

    def _monitor_progress(self, process, progress_callback):
        """
        Monitor FFmpeg progress and call the callback function.

        Args:
            process: Subprocess process
            progress_callback: Function to call with progress (0-1)

        Returns:
            Tuple of (stdout, stderr)
        """
        stdout_data = b""
        stderr_data = b""

        # Read output in a non-blocking way
        while process.poll() is None:
            # Read stdout
            if process.stdout:
                line = process.stdout.readline()
                if line:
                    stdout_data += line
                    # Parse progress information
                    try:
                        line_str = line.decode("utf-8", errors="replace").strip()
                        if line_str.startswith("out_time_ms="):
                            time_ms = int(line_str.split("=")[1])
                            progress = time_ms / 1000 / 60  # Approximate progress
                            progress_callback(min(1.0, max(0.0, progress)))
                    except Exception as e:
                        logger.debug(f"Error parsing progress: {e}")

            # Read stderr
            if process.stderr:
                line = process.stderr.readline()
                if line:
                    stderr_data += line

        # Read any remaining output
        if process.stdout:
            stdout_data += process.stdout.read()
        if process.stderr:
            stderr_data += process.stderr.read()

        return stdout_data, stderr_data

    def _run_ffprobe(self, args: List[str]) -> Dict[str, Any]:
        """
        Run FFprobe with the given arguments.

        Args:
            args: FFprobe command-line arguments

        Returns:
            Parsed JSON output

        Raises:
            RuntimeError: If FFprobe returns a non-zero exit code
        """
        # Add FFprobe executable and JSON output format to the arguments
        cmd = [
            self._ffprobe_path,
            "-v", "quiet",
            "-print_format", "json",
            "-show_format",
            "-show_streams"
        ] + args

        # Log the command
        logger.debug(f"Running FFprobe: {' '.join(cmd)}")

        # Run FFprobe
        process = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=False
        )

        stdout, stderr = process.communicate()

        # Check return code
        if process.returncode != 0:
            error_msg = stderr.decode("utf-8", errors="replace")
            logger.error(f"FFprobe error (code {process.returncode}): {error_msg}")
            raise RuntimeError(f"FFprobe error (code {process.returncode}): {error_msg}")

        # Parse JSON output
        try:
            result = json.loads(stdout.decode("utf-8"))
        except json.JSONDecodeError:
            logger.error("Failed to parse FFprobe JSON output")
            result = {}

        return result

    def get_info(self, video: Dict[str, Any]) -> Dict[str, Any]:
        """
        Get video information.

        Args:
            video: Video dictionary

        Returns:
            Dictionary with video info
        """
        # If info is already in the video dictionary and is processed, return it
        if "info" in video and isinstance(video["info"], dict) and "width" in video["info"]:
            return video["info"]

        # Run FFprobe to get video info if not already available
        if "info" in video:
            info = video["info"]
        else:
            info = self._run_ffprobe(["-i", video["path"]])

        # Find video stream
        video_stream = None
        for stream in info.get("streams", []):
            if stream.get("codec_type") == "video":
                video_stream = stream
                break

        # Find audio stream
        audio_stream = None
        for stream in info.get("streams", []):
            if stream.get("codec_type") == "audio":
                audio_stream = stream
                break

        # Extract video information
        width = int(video_stream.get("width", 0)) if video_stream else 0
        height = int(video_stream.get("height", 0)) if video_stream else 0

        # Extract duration
        duration = float(info.get("format", {}).get("duration", 0))

        # Apply start and end times if specified
        if video.get("start", 0) > 0:
            duration -= video["start"]

        if video.get("end") is not None:
            duration = min(duration, video["end"] - video.get("start", 0))

        # Extract FPS
        fps = 0
        if video_stream and "avg_frame_rate" in video_stream:
            try:
                fps_fraction = video_stream["avg_frame_rate"].split("/")
                fps = float(fps_fraction[0]) / float(fps_fraction[1])
            except (ValueError, ZeroDivisionError, IndexError):
                fps = 0

        # Extract codec information
        video_codec = video_stream.get("codec_name", "unknown") if video_stream else "none"
        audio_codec = audio_stream.get("codec_name", "unknown") if audio_stream else "none"

        # Extract bitrate
        bitrate = info.get("format", {}).get("bit_rate")
        if bitrate:
            try:
                bitrate = f"{int(bitrate) // 1000}k"
            except (ValueError, TypeError):
                bitrate = "unknown"

        return {
            "width": width,
            "height": height,
            "duration": duration,
            "fps": fps,
            "has_audio": audio_stream is not None,
            "codec": video_codec,
            "audio_codec": audio_codec if audio_stream else None,
            "bitrate": bitrate
        }



#!/usr/bin/env python
"""
Example of AI-powered features in FMUS-VID.

This script demonstrates object detection, tracking, and background removal.
"""

import sys
import os
import time
from pathlib import Path
import cv2
import numpy as np

# Add the parent directory to the path so we can import fmusvid
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from fmusvid import Video
from fmusvid.ai.detection import ObjectDetector
from fmusvid.ai.tracking import ObjectTracker
from fmusvid.ai.background import BackgroundRemover, BackgroundBlur

def run_object_detection(video_path):
    """Run object detection on a video file."""
    print(f"Running object detection on {video_path}")

    try:
        # Load video
        cap = cv2.VideoCapture(str(video_path))
        if not cap.isOpened():
            print(f"Error: Could not open video {video_path}")
            return None

        # Get video properties
        fps = cap.get(cv2.CAP_PROP_FPS)
        width = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
        height = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))

        # Create output video
        output_dir = os.path.dirname(video_path)
        output_filename = f"{os.path.splitext(os.path.basename(video_path))[0]}_detected.mp4"
        output_path = os.path.join(output_dir, output_filename)

        out = cv2.VideoWriter(
            str(output_path),
            cv2.VideoWriter_fourcc(*'mp4v'),
            fps,
            (width, height)
        )

        # Initialize object detector
        try:
            detector = ObjectDetector()
            print(f"Loaded {detector.model_name} model on {detector.device}")
            has_detector = True
        except Exception as e:
            print(f"Error initializing object detector: {e}")
            print("Falling back to drawing placeholders instead")
            has_detector = False

        # Process video
        frame_count = 0
        start_time = time.time()

        while True:
            ret, frame = cap.read()
            if not ret:
                break

            # Run detection or create placeholder visualization
            if has_detector:
                try:
                    # Detect objects
                    detections = detector.detect(frame, conf_threshold=0.5)

                    # Annotate frame
                    result = detector.annotate_frame(frame, detections)

                    # Add detection count
                    cv2.putText(
                        result,
                        f"Detected: {len(detections)} objects",
                        (10, 30),
                        cv2.FONT_HERSHEY_SIMPLEX,
                        1.0,
                        (0, 255, 0),
                        2
                    )
                except Exception as e:
                    print(f"Error running detection: {e}")
                    result = frame.copy()  # Use original frame if detection fails

                    # Add error message
                    cv2.putText(
                        result,
                        "Detection error",
                        (10, 30),
                        cv2.FONT_HERSHEY_SIMPLEX,
                        1.0,
                        (0, 0, 255),
                        2
                    )
            else:
                # Create a placeholder visualization
                result = frame.copy()

                # Draw some random detection boxes as placeholders
                for i in range(3):
                    x = np.random.randint(10, width - 100)
                    y = np.random.randint(10, height - 100)
                    w = np.random.randint(50, 200)
                    h = np.random.randint(50, 200)

                    # Ensure box is within frame
                    if x + w >= width: w = width - x - 1
                    if y + h >= height: h = height - y - 1

                    # Draw box
                    cv2.rectangle(result, (x, y), (x + w, y + h), (0, 255, 0), 2)

                    # Draw label
                    labels = ["Person", "Car", "Dog", "Cat", "Bird"]
                    label = labels[i % len(labels)]
                    cv2.putText(
                        result,
                        label,
                        (x, y - 10),
                        cv2.FONT_HERSHEY_SIMPLEX,
                        0.5,
                        (0, 255, 0),
                        1
                    )

                # Add simulation message
                cv2.putText(
                    result,
                    "Simulated detection (no model available)",
                    (10, 30),
                    cv2.FONT_HERSHEY_SIMPLEX,
                    0.7,
                    (0, 255, 255),
                    2
                )

            # Write frame
            out.write(result)

            # Update progress
            frame_count += 1
            if frame_count % 10 == 0:
                elapsed = time.time() - start_time
                print(f"Processed {frame_count} frames in {elapsed:.2f}s ({frame_count/elapsed:.2f} fps)")

        # Clean up
        cap.release()
        out.release()
        print(f"Saved detection results to {output_path}")
        return output_path

    except Exception as e:
        print(f"Error in object detection: {e}")
        return None

def run_object_tracking(video_path):
    """Run object tracking on a video file."""
    print(f"Running object tracking on {video_path}")

    try:
        # Load video
        cap = cv2.VideoCapture(str(video_path))
        if not cap.isOpened():
            print(f"Error: Could not open video {video_path}")
            return None

        # Get video properties
        fps = cap.get(cv2.CAP_PROP_FPS)
        width = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
        height = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))

        # Create output video
        output_dir = os.path.dirname(video_path)
        output_filename = f"{os.path.splitext(os.path.basename(video_path))[0]}_tracked.mp4"
        output_path = os.path.join(output_dir, output_filename)

        out = cv2.VideoWriter(
            str(output_path),
            cv2.VideoWriter_fourcc(*'mp4v'),
            fps,
            (width, height)
        )

        # Initialize object detector and tracker
        try:
            detector = ObjectDetector()
            tracker = ObjectTracker()
            print(f"Loaded {detector.model_name} model on {detector.device}")
            has_tracking = True
        except Exception as e:
            print(f"Error initializing object detector or tracker: {e}")
            print("Falling back to drawing placeholders instead")
            has_tracking = False

        # Process video
        frame_count = 0
        start_time = time.time()

        # For placeholder tracking
        placeholder_tracks = []

        while True:
            ret, frame = cap.read()
            if not ret:
                break

            # Run tracking or create placeholder visualization
            if has_tracking:
                try:
                    # Only run detection every 5 frames for performance
                    if frame_count % 5 == 0:
                        detections = detector.detect(frame, conf_threshold=0.5)
                    else:
                        detections = []

                    # Update tracker
                    tracks = tracker.update(detections)

                    # Annotate frame
                    result = tracker.annotate_frame(frame)

                    # Add tracking count
                    cv2.putText(
                        result,
                        f"Tracking: {len(tracks)} objects",
                        (10, 30),
                        cv2.FONT_HERSHEY_SIMPLEX,
                        1.0,
                        (0, 255, 0),
                        2
                    )
                except Exception as e:
                    print(f"Error running tracking: {e}")
                    result = frame.copy()  # Use original frame if tracking fails

                    # Add error message
                    cv2.putText(
                        result,
                        "Tracking error",
                        (10, 30),
                        cv2.FONT_HERSHEY_SIMPLEX,
                        1.0,
                        (0, 0, 255),
                        2
                    )
            else:
                # Create a placeholder visualization with simulated tracking
                result = frame.copy()

                # Initialize placeholder tracks if needed
                if frame_count == 0 or len(placeholder_tracks) == 0:
                    placeholder_tracks = []
                    for i in range(3):
                        x = np.random.randint(10, width - 100)
                        y = np.random.randint(10, height - 100)
                        w = np.random.randint(50, 150)
                        h = np.random.randint(50, 150)
                        dx = np.random.randint(-5, 5)
                        dy = np.random.randint(-5, 5)

                        # Ensure box is within frame
                        if x + w >= width: w = width - x - 1
                        if y + h >= height: h = height - y - 1

                        # Create track
                        placeholder_tracks.append({
                            "id": i + 1,
                            "x": x,
                            "y": y,
                            "w": w,
                            "h": h,
                            "dx": dx,
                            "dy": dy,
                            "label": ["Person", "Car", "Dog"][i % 3]
                        })

                # Update and draw placeholder tracks
                for track in placeholder_tracks:
                    # Update position
                    track["x"] += track["dx"]
                    track["y"] += track["dy"]

                    # Bounce off edges
                    if track["x"] < 0 or track["x"] + track["w"] >= width:
                        track["dx"] *= -1
                        track["x"] = max(0, min(width - track["w"], track["x"]))

                    if track["y"] < 0 or track["y"] + track["h"] >= height:
                        track["dy"] *= -1
                        track["y"] = max(0, min(height - track["h"], track["y"]))

                    # Draw track
                    cv2.rectangle(
                        result,
                        (int(track["x"]), int(track["y"])),
                        (int(track["x"] + track["w"]), int(track["y"] + track["h"])),
                        (0, 255, 255),
                        2
                    )

                    # Draw ID and label
                    cv2.putText(
                        result,
                        f"ID: {track['id']} ({track['label']})",
                        (int(track["x"]), int(track["y"]) - 10),
                        cv2.FONT_HERSHEY_SIMPLEX,
                        0.5,
                        (0, 255, 255),
                        1
                    )

                # Add simulation message
                cv2.putText(
                    result,
                    "Simulated tracking (no model available)",
                    (10, 30),
                    cv2.FONT_HERSHEY_SIMPLEX,
                    0.7,
                    (0, 255, 255),
                    2
                )

            # Write frame
            out.write(result)

            # Update progress
            frame_count += 1
            if frame_count % 10 == 0:
                elapsed = time.time() - start_time
                print(f"Processed {frame_count} frames in {elapsed:.2f}s ({frame_count/elapsed:.2f} fps)")

        # Clean up
        cap.release()
        out.release()
        print(f"Saved tracking results to {output_path}")
        return output_path

    except Exception as e:
        print(f"Error in object tracking: {e}")
        return None

def run_background_removal(video_path, background_path=None):
    """Run background removal or replacement on a video file."""
    print(f"Running background removal on {video_path}")

    try:
        # Load video
        cap = cv2.VideoCapture(str(video_path))
        if not cap.isOpened():
            print(f"Error: Could not open video {video_path}")
            return None

        # Get video properties
        fps = cap.get(cv2.CAP_PROP_FPS)
        width = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
        height = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))

        # Load background if provided
        background_img = None
        if background_path and os.path.exists(background_path):
            try:
                background_img = cv2.imread(str(background_path))
                if background_img is not None:
                    background_img = cv2.cvtColor(background_img, cv2.COLOR_BGR2RGB)
                    background_img = cv2.resize(background_img, (width, height))
                    print(f"Loaded background image: {background_path}")
                else:
                    print(f"Failed to load background image: {background_path}")
                    background_img = None
            except Exception as e:
                print(f"Error loading background image: {e}")
                background_img = None

        # Determine output format
        output_dir = os.path.dirname(video_path)
        if background_img is not None:
            output_filename = f"{os.path.splitext(os.path.basename(video_path))[0]}_bg_replaced.mp4"
        else:
            output_filename = f"{os.path.splitext(os.path.basename(video_path))[0]}_bg_removed.mp4"

        output_path = os.path.join(output_dir, output_filename)
        fourcc = cv2.VideoWriter_fourcc(*'mp4v')
        out = cv2.VideoWriter(str(output_path), fourcc, fps, (width, height))

        # Initialize background remover
        try:
            if background_img is not None:
                bg_processor = BackgroundRemover()
            else:
                bg_processor = BackgroundBlur()
            print(f"Loaded {bg_processor.model_name if hasattr(bg_processor, 'model_name') else 'background blur'} model")
            has_bg_processor = True
        except Exception as e:
            print(f"Error initializing background remover: {e}")
            print("Falling back to simple background processing")
            has_bg_processor = False

        # Process video
        frame_count = 0
        start_time = time.time()

        while True:
            ret, frame = cap.read()
            if not ret:
                break

            # Process frame
            try:
                if has_bg_processor:
                    # Apply background removal/replacement
                    if background_img is not None:
                        # With custom background
                        result = bg_processor.process(frame, background=background_img)
                    else:
                        # With blur
                        result = bg_processor.process(frame)
                else:
                    # Create a simple effect if models are not available
                    result = frame.copy()

                    # Create a simple mask (circular gradient from center) - vectorized
                    y_coords, x_coords = np.meshgrid(np.arange(height), np.arange(width), indexing='ij')
                    center_x, center_y = width // 2, height // 2
                    dist = np.sqrt((x_coords - center_x)**2 + (y_coords - center_y)**2) / np.sqrt((width//2)**2 + (height//2)**2)
                    mask = np.maximum(0, 1 - dist).astype(np.float32)

                    # Apply simple foreground/background processing
                    if background_img is not None:
                        # Apply mask with background
                        mask_3d = np.repeat(mask[:, :, np.newaxis], 3, axis=2)
                        result = (frame * mask_3d + background_img * (1 - mask_3d)).astype(np.uint8)
                    else:
                        # Apply blur effect
                        blurred = cv2.GaussianBlur(frame, (25, 25), 0)
                        mask_3d = np.repeat(mask[:, :, np.newaxis], 3, axis=2)
                        result = (frame * mask_3d + blurred * (1 - mask_3d)).astype(np.uint8)

                    # Add simulation message
                    cv2.putText(
                        result,
                        "Simulated background effect (no model available)",
                        (10, 30),
                        cv2.FONT_HERSHEY_SIMPLEX,
                        0.7,
                        (0, 255, 255),
                        2
                    )
            except Exception as e:
                print(f"Error processing frame {frame_count}: {e}")
                result = frame.copy()

                # Add error message
                cv2.putText(
                    result,
                    "Processing error",
                    (10, 30),
                    cv2.FONT_HERSHEY_SIMPLEX,
                    1.0,
                    (0, 0, 255),
                    2
                )

            # Write frame
            out.write(cv2.cvtColor(result, cv2.COLOR_RGB2BGR))

            # Update progress
            frame_count += 1
            if frame_count % 10 == 0:
                elapsed = time.time() - start_time
                print(f"Processed {frame_count} frames in {elapsed:.2f}s ({frame_count/elapsed:.2f} fps)")

        # Clean up
        cap.release()
        out.release()
        print(f"Saved background processing results to {output_path}")
        return output_path

    except Exception as e:
        print(f"Error in background removal: {e}")
        return None

def main():
    """Run the AI features example."""
    print("FMUS-VID AI Features Example")
    print("============================")

    # Hardcoded sample videos to try
    sample_video_paths = [
        os.path.join("test_media", "test_video.mp4"),
        os.path.join("test_media", "sample.mp4"),
        os.path.join("test_media", "sample2.mp4")
    ]

    # Find the first existing sample video
    video_path = None
    for path in sample_video_paths:
        if os.path.exists(path):
            video_path = path
            break

    # If no sample videos found, create a simple test video
    if video_path is None:
        print("No sample videos found. Creating a simple test video...")

        # Create test_media directory if it doesn't exist
        os.makedirs("test_media", exist_ok=True)

        # Create a simple test video with colored rectangles moving
        test_video_path = os.path.join("test_media", "generated_test.mp4")

        # Video properties
        width, height = 640, 480
        fps = 30
        duration = 5  # seconds

        # Create video writer
        fourcc = cv2.VideoWriter_fourcc(*'mp4v')
        out = cv2.VideoWriter(test_video_path, fourcc, fps, (width, height))

        # Generate frames with moving rectangles
        for i in range(int(fps * duration)):
            # Create black frame
            frame = np.zeros((height, width, 3), dtype=np.uint8)

            # Draw moving rectangles
            cv2.rectangle(frame, (i*10 % (width-100), 100), (i*10 % (width-100) + 100, 200), (0, 0, 255), -1)
            cv2.rectangle(frame, (width - (i*8 % (width-80)), 300), (width - (i*8 % (width-80)) + 80, 380), (0, 255, 0), -1)

            # Write frame
            out.write(frame)

        out.release()
        print(f"Created test video: {test_video_path}")
        video_path = test_video_path

    print(f"\nUsing video: {video_path}")

    # Run all AI features
    print("\nRunning Object Detection...")
    run_object_detection(video_path)

    print("\nRunning Object Tracking...")
    run_object_tracking(video_path)

    print("\nRunning Background Blur...")
    run_background_removal(video_path)

    # Generate a simple background image for background replacement
    background_path = os.path.join("test_media", "background.jpg")
    if not os.path.exists(background_path):
        print(f"Creating a simple background image: {background_path}")
        bg_img = np.ones((480, 640, 3), dtype=np.uint8) * 255  # White background

        # Add a simple gradient (vectorized)
        y_coords = np.arange(480)
        gradient = np.stack([
            np.zeros(480, dtype=np.uint8),
            (y_coords / 480 * 255).astype(np.uint8),
            ((1 - y_coords / 480) * 255).astype(np.uint8)
        ], axis=1)
        bg_img = np.tile(gradient[:, np.newaxis, :], (1, 640, 1))

        cv2.imwrite(background_path, cv2.cvtColor(bg_img, cv2.COLOR_RGB2BGR))

    print("\nRunning Background Replacement...")
    run_background_removal(video_path, background_path)

    print("\nAll AI features have been demonstrated.")
    print("Check the output files in the same directory as the input video.")

if __name__ == "__main__":
    main()



"""
Subtitle support for FMUS-VID.

This module provides functionality for adding and manipulating subtitles.
"""

from typing import Union, List, Dict, Optional, Tuple
from pathlib import Path
import re
import logging
from dataclasses import dataclass
from datetime import datetime, timedelta
from PIL import Image, ImageDraw, ImageFont, ImageColor

logger = logging.getLogger(__name__)

@dataclass
class SubtitleEntry:
    """Represents a single subtitle entry."""
    start_time: float  # Start time in seconds
    end_time: float    # End time in seconds
    text: str         # Subtitle text
    position: Optional[Tuple[int, int]] = None  # Optional position override

class SubtitleParser:
    """Parser for various subtitle formats."""

    @staticmethod
    def parse(path: Union[str, Path], format: Optional[str] = None) -> List[SubtitleEntry]:
        """
        Parse a subtitle file.

        Args:
            path: Path to subtitle file
            format: Subtitle format ("srt", "vtt", etc.) or None for auto-detect

        Returns:
            List of SubtitleEntry objects
        """
        path = Path(path)

        # Auto-detect format from extension if not specified
        if format is None:
            format = path.suffix.lower().lstrip('.')

        # Select parser based on format
        if format == "srt":
            return SubtitleParser._parse_srt(path)
        elif format == "vtt":
            return SubtitleParser._parse_vtt(path)
        else:
            raise ValueError(f"Unsupported subtitle format: {format}")

    @staticmethod
    def _parse_srt(path: Path) -> List[SubtitleEntry]:
        """Parse an SRT subtitle file."""
        entries = []
        current_entry = None

        with open(path, 'r', encoding='utf-8-sig') as f:
            lines = f.readlines()

        for line in lines:
            line = line.strip()

            if not line:
                if current_entry:
                    entries.append(current_entry)
                    current_entry = None
                continue

            # Parse index number
            if current_entry is None:
                try:
                    int(line)
                    current_entry = {"text": []}
                    continue
                except ValueError:
                    pass

            # Parse timestamp line
            if current_entry and not current_entry.get("start_time"):
                match = re.match(r'(\d{2}:\d{2}:\d{2},\d{3}) --> (\d{2}:\d{2}:\d{2},\d{3})', line)
                if match:
                    current_entry["start_time"] = SubtitleParser._parse_timestamp(match.group(1))
                    current_entry["end_time"] = SubtitleParser._parse_timestamp(match.group(2))
                    continue

            # Add text line
            if current_entry:
                current_entry["text"].append(line)

        # Add last entry if exists
        if current_entry:
            entries.append(current_entry)

        # Convert to SubtitleEntry objects
        return [
            SubtitleEntry(
                start_time=entry["start_time"],
                end_time=entry["end_time"],
                text="\n".join(entry["text"])
            )
            for entry in entries
        ]

    @staticmethod
    def _parse_vtt(path: Path) -> List[SubtitleEntry]:
        """Parse a WebVTT subtitle file."""
        entries = []
        current_entry = None
        header_passed = False

        with open(path, 'r', encoding='utf-8-sig') as f:
            lines = f.readlines()

        for line in lines:
            line = line.strip()

            # Skip WebVTT header
            if not header_passed:
                if line.startswith('WEBVTT'):
                    header_passed = True
                continue

            if not line:
                if current_entry:
                    entries.append(current_entry)
                    current_entry = None
                continue

            # Parse timestamp line
            match = re.match(r'(\d{2}:\d{2}:\d{2}\.\d{3}) --> (\d{2}:\d{2}:\d{2}\.\d{3})', line)
            if match:
                current_entry = {
                    "start_time": SubtitleParser._parse_timestamp(match.group(1)),
                    "end_time": SubtitleParser._parse_timestamp(match.group(2)),
                    "text": []
                }
                continue

            # Add text line
            if current_entry:
                current_entry["text"].append(line)

        # Add last entry if exists
        if current_entry:
            entries.append(current_entry)

        # Convert to SubtitleEntry objects
        return [
            SubtitleEntry(
                start_time=entry["start_time"],
                end_time=entry["end_time"],
                text="\n".join(entry["text"])
            )
            for entry in entries
        ]

    @staticmethod
    def _parse_timestamp(timestamp: str) -> float:
        """Convert timestamp string to seconds."""
        # Handle both SRT (,) and VTT (.) decimal separators
        timestamp = timestamp.replace(',', '.')

        # Parse timestamp
        time_obj = datetime.strptime(timestamp, '%H:%M:%S.%f')
        return (time_obj.hour * 3600 +
                time_obj.minute * 60 +
                time_obj.second +
                time_obj.microsecond / 1_000_000)

class SubtitleRenderer:
    """Renders subtitles on video frames."""

    def __init__(self, entries: List[SubtitleEntry],
                 font: str = "Arial", size: int = 24,
                 color: Union[str, Tuple[int, int, int]] = "white",
                 stroke_width: int = 2,
                 stroke_color: Union[str, Tuple[int, int, int]] = "black",
                 position: Optional[Tuple[int, int]] = None,
                 fade_in: float = 0.25, fade_out: float = 0.25):
        """
        Initialize subtitle renderer.

        Args:
            entries: List of subtitle entries
            font: Font name or path
            size: Font size in pixels
            color: Text color
            stroke_width: Outline width (0 for no outline)
            stroke_color: Outline color
            position: Default (x, y) position (None for bottom center)
            fade_in: Fade-in duration in seconds
            fade_out: Fade-out duration in seconds
        """
        self.entries = entries
        self.font = font
        self.size = size
        self.color = color
        self.stroke_width = stroke_width
        self.stroke_color = stroke_color
        self.default_position = position
        self.fade_in = fade_in
        self.fade_out = fade_out

        # Try to load the font
        try:
            self._font_obj = ImageFont.truetype(font, size)
        except IOError:
            logger.warning(f"Font '{font}' not found, using default font")
            self._font_obj = ImageFont.load_default()

    def render(self, frame: Image.Image, time: float) -> Image.Image:
        """
        Render subtitles for the current frame.

        Args:
            frame: PIL Image to render on
            time: Current video time in seconds

        Returns:
            Frame with rendered subtitles
        """
        # Find visible subtitles at current time
        visible_entries = [
            entry for entry in self.entries
            if entry.start_time <= time <= entry.end_time
        ]

        if not visible_entries:
            return frame

        # Create a copy of the frame
        result = frame.copy()

        # Calculate default position if not specified
        if self.default_position is None:
            width, height = frame.size
            x = width // 2
            y = height - 50  # 50 pixels from bottom
            self.default_position = (x, y)

        # Render each visible subtitle
        for entry in visible_entries:
            # Calculate opacity based on fade effects
            opacity = 1.0

            if time < entry.start_time + self.fade_in:
                opacity = (time - entry.start_time) / self.fade_in
            elif time > entry.end_time - self.fade_out:
                opacity = (entry.end_time - time) / self.fade_out

            opacity = max(0.0, min(1.0, opacity))

            if opacity == 0:
                continue

            # Use entry-specific position or default
            position = entry.position or self.default_position

            # Create a temporary image for the subtitle
            subtitle_overlay = Image.new('RGBA', frame.size, (0, 0, 0, 0))
            draw = ImageDraw.Draw(subtitle_overlay)

            # Draw text with stroke if specified
            if self.stroke_width > 0:
                # Draw stroke
                for dx in range(-self.stroke_width, self.stroke_width + 1):
                    for dy in range(-self.stroke_width, self.stroke_width + 1):
                        draw.text(
                            (position[0] + dx, position[1] + dy),
                            entry.text,
                            font=self._font_obj,
                            fill=(*self._get_color(self.stroke_color), int(255 * opacity)),
                            align="center"
                        )

            # Draw main text
            draw.text(
                position,
                entry.text,
                font=self._font_obj,
                fill=(*self._get_color(self.color), int(255 * opacity)),
                align="center"
            )

            # Composite the subtitle overlay onto the frame
            result = Image.alpha_composite(result.convert('RGBA'), subtitle_overlay)

        return result

    def _get_color(self, color: Union[str, Tuple[int, int, int]]) -> Tuple[int, int, int]:
        """Convert color to RGB tuple."""
        if isinstance(color, tuple):
            return color
        return ImageColor.getrgb(color)



"""
Tests for video composition functionality.
"""

import pytest
import fmusvid
from pathlib import Path

@pytest.fixture
def sample_videos(tmp_path):
    """Create sample videos for testing."""
    videos = []
    for i in range(4):
        path = tmp_path / f"clip{i}.mp4"
        video = fmusvid.create(
            width=640,
            height=360,
            duration=5.0,
            fps=30
        )
        video.save(path)
        videos.append(path)
    return videos

def test_picture_in_picture(sample_videos):
    """Test picture-in-picture overlay."""
    main = fmusvid.load(sample_videos[0])
    overlay = fmusvid.load(sample_videos[1])

    result = main.overlay(
        overlay.resize(width=320),
        position=(50, 50),
        start=1.0,
        duration=3.0
    )

    assert result.width == main.width
    assert result.height == main.height
    assert result.duration == main.duration

def test_video_grid(sample_videos):
    """Test video grid creation."""
    videos = [fmusvid.load(path) for path in sample_videos]

    grid = fmusvid.grid(
        videos,
        rows=2,
        cols=2
    )

    # Grid scales videos to fit: width is maintained, height is divided by rows
    expected_width = videos[0].width * 2
    expected_height = (videos[0].height // 2) * 2

    assert grid.width == expected_width
    assert grid.height == expected_height
    assert grid.duration == max(v.duration for v in videos)

def test_transitions(sample_videos):
    """Test video transitions."""
    pytest.skip("transition module not yet exposed in API")

def test_split_screen(sample_videos):
    """Test split screen composition."""
    pytest.skip("compose function not yet implemented")



"""
Backend manager for FMUS-VID.

This module manages the available video processing backends.
"""

import logging
from enum import Enum, auto
from typing import Union, Type, Dict, Any, Optional, Tuple, Callable, List
from pathlib import Path

from .backend import Backend as BackendInterface

logger = logging.getLogger(__name__)

class Backend(Enum):
    """Enum of available backends."""

    FFMPEG = auto()
    MOVIEPY = auto()
    PYAV = auto()
    GSTREAMER = auto()
    VLC = auto()
    DUMMY = auto()  # Added dummy backend for testing/CLI

    @staticmethod
    def from_string(name: str) -> 'Backend':
        """
        Convert a string to a Backend enum.

        Args:
            name: Backend name (case-insensitive)

        Returns:
            Backend enum

        Raises:
            ValueError: If the backend is not recognized
        """
        name = name.upper()

        if name == "FFMPEG":
            return Backend.FFMPEG
        elif name == "MOVIEPY":
            return Backend.MOVIEPY
        elif name == "PYAV":
            return Backend.PYAV
        elif name == "GSTREAMER":
            return Backend.GSTREAMER
        elif name == "VLC":
            return Backend.VLC
        elif name == "DUMMY":
            return Backend.DUMMY
        else:
            raise ValueError(f"Unknown backend: {name}")


# Dummy backend implementation
class DummyBackend(BackendInterface):
    """
    Dummy backend for testing and CLI functionality when no real backends are available.
    """

    @staticmethod
    def is_available() -> bool:
        """Always available."""
        return True

    def load(self, path: Union[str, Path], **kwargs) -> Any:
        """Pretend to load a video."""
        logger.warning("Using dummy backend - no real video operations will be performed")
        return {"path": path, "loaded": True}

    def create(self, width: int, height: int, duration: float, fps: float,
               color: Tuple[int, int, int] = (0, 0, 0), **kwargs) -> Any:
        """Create a blank video."""
        logger.info(f"Dummy create: {width}x{height}, {duration}s at {fps} fps")
        return {"width": width, "height": height, "duration": duration, "fps": fps, "color": color}

    def save(self, video: Any, output_path: Union[str, Path],
             progress_callback: Optional[Callable[[float], None]] = None, **kwargs) -> None:
        """Pretend to save a video."""
        logger.info(f"Dummy save to {output_path}")
        if progress_callback:
            progress_callback(1.0)  # Simulate completion

    def get_info(self, video: Any) -> Dict[str, Any]:
        """Return dummy video info."""
        return {
            "width": 1920,
            "height": 1080,
            "duration": 30.0,
            "fps": 30.0,
            "has_audio": True,
            "codec": "h264",
            "audio_codec": "aac",
            "bitrate": "5M"
        }

    # Dummy implementation of all basic operations
    def trim(self, video: Any, start: float, end: Optional[float] = None, **kwargs) -> Any:
        logger.info(f"Dummy trim: {start} to {end}")
        return video

    def resize(self, video: Any, width: Optional[int] = None, height: Optional[int] = None,
              keep_aspect: bool = True, **kwargs) -> Any:
        logger.info(f"Dummy resize: {width}x{height}, keep_aspect={keep_aspect}")
        return video

    def crop(self, video: Any, x: int, y: int, width: int, height: int, **kwargs) -> Any:
        logger.info(f"Dummy crop: ({x}, {y}), {width}x{height}")
        return video

    def rotate(self, video: Any, degrees: float, **kwargs) -> Any:
        logger.info(f"Dummy rotate: {degrees}°")
        return video

    def grayscale(self, video: Any, **kwargs) -> Any:
        logger.info("Dummy grayscale")
        return video

    def blur(self, video: Any, radius: float, **kwargs) -> Any:
        logger.info(f"Dummy blur: radius={radius}")
        return video

    def brightness(self, video: Any, factor: float, **kwargs) -> Any:
        logger.info(f"Dummy brightness: factor={factor}")
        return video

    def contrast(self, video: Any, factor: float, **kwargs) -> Any:
        logger.info(f"Dummy contrast: factor={factor}")
        return video

    def mute(self, video: Any, **kwargs) -> Any:
        logger.info("Dummy mute")
        return video

    def volume(self, video: Any, level: float, **kwargs) -> Any:
        logger.info(f"Dummy volume: level={level}")
        return video

    def add_audio(self, video: Any, audio_path: Union[str, Path],
                start: float = 0, volume: float = 1.0, **kwargs) -> Any:
        logger.info(f"Dummy add_audio: {audio_path}, start={start}, volume={volume}")
        return video

    def extract_frame(self, video: Any, time: float, **kwargs) -> Any:
        logger.info(f"Dummy extract_frame at time {time}")
        return {"frame": True, "time": time}

    def overlay(self, video: Any, overlay_video: Any, position=(0, 0),
               start: float = 0, duration: Optional[float] = None, opacity: float = 1.0, **kwargs) -> Any:
        logger.info(f"Dummy overlay: position={position}, start={start}, duration={duration}, opacity={opacity}")
        return video

    def concat(self, videos: List[Any], **kwargs) -> Any:
        logger.info(f"Dummy concat {len(videos)} videos")
        return videos[0] if videos else None

    def grid(self, videos: List[Any], rows: int, cols: int, **kwargs) -> Any:
        logger.info(f"Dummy grid: {rows}x{cols} with {len(videos)} videos")
        return videos[0] if videos else None

    def add_subtitles(self, video: Any, entries: List,
                     font: str = "Arial", size: int = 24,
                     color: Union[str, Tuple[int, int, int]] = "white",
                     position: Optional[Tuple[int, int]] = None, **kwargs) -> Any:
        logger.info(f"Dummy add_subtitles: {len(entries)} entries, font={font}, size={size}")
        return video

    def add_subtitle_text(self, video: Any, entry,
                         font: str = "Arial", size: int = 24,
                         color: Union[str, Tuple[int, int, int]] = "white", **kwargs) -> Any:
        logger.info(f"Dummy add_subtitle_text: font={font}, size={size}")
        return video


def get_backend(backend_name: Union[str, Backend] = "auto") -> BackendInterface:
    """
    Get a video processing backend.

    Args:
        backend_name: Backend name or enum ("auto", "ffmpeg", "moviepy", "pyav", "gstreamer", "vlc")

    Returns:
        Backend instance

    Raises:
        ValueError: If the backend is not available
    """
    # Check if a Backend enum was passed
    if isinstance(backend_name, Backend):
        backend_enum = backend_name
    elif backend_name == "auto":
        # Auto-detect the best available backend
        backend_enum = _auto_detect_backend()
    else:
        # Convert string to enum
        backend_enum = Backend.from_string(backend_name)

    # Import and create the appropriate backend
    try:
        if backend_enum == Backend.FFMPEG:
            # Use the new ffmpeg package
            from .ffmpeg import FFmpegBackend
            backend_instance = FFmpegBackend()
        elif backend_enum == Backend.MOVIEPY:
            from .moviepy import MoviePyBackend
            backend_instance = MoviePyBackend()
        elif backend_enum == Backend.PYAV:
            from .pyav import PyAVBackend
            backend_instance = PyAVBackend()
        elif backend_enum == Backend.GSTREAMER:
            from .gstreamer import GStreamerBackend
            backend_instance = GStreamerBackend()
        elif backend_enum == Backend.VLC:
            from .vlc import VLCBackend
            backend_instance = VLCBackend()
        elif backend_enum == Backend.DUMMY:
            backend_instance = DummyBackend()
        else:
            raise ValueError(f"Unknown backend: {backend_enum}")
    except (ImportError, AttributeError, TypeError) as e:
        logger.warning(f"Error importing backend {backend_enum.name}: {e}. Falling back to dummy backend.")
        backend_instance = DummyBackend()

    return backend_instance

def _auto_detect_backend() -> Backend:
    """
    Auto-detect the best available backend.

    Returns:
        Backend enum

    Raises:
        ValueError: If no backend is available
    """
    # Try backends in order of preference
    backends = [
        (Backend.FFMPEG, lambda: _check_backend_available(Backend.FFMPEG)),
        (Backend.PYAV, lambda: _check_backend_available(Backend.PYAV)),
        (Backend.MOVIEPY, lambda: _check_backend_available(Backend.MOVIEPY)),
        (Backend.GSTREAMER, lambda: _check_backend_available(Backend.GSTREAMER)),
        (Backend.VLC, lambda: _check_backend_available(Backend.VLC))
    ]

    for backend_enum, check_func in backends:
        try:
            if check_func():
                logger.info(f"Auto-detected backend: {backend_enum.name}")
                return backend_enum
        except Exception as e:
            logger.debug(f"Error checking {backend_enum.name} availability: {e}")

    # If we get here, no backend is available - use dummy backend
    logger.warning("No video processing backends available. Using dummy backend instead.")
    return Backend.DUMMY

def _check_backend_available(backend_enum: Backend) -> bool:
    """
    Check if a backend is available.

    Args:
        backend_enum: Backend enum

    Returns:
        True if the backend is available, False otherwise
    """
    if backend_enum == Backend.FFMPEG:
        try:
            # Use the new ffmpeg package
            from .ffmpeg import FFmpegBackend
            return FFmpegBackend.is_available()
        except (ImportError, AttributeError, TypeError):
            logger.debug("FFMPEG backend not available")
            return False
    elif backend_enum == Backend.MOVIEPY:
        try:
            from .moviepy import MoviePyBackend
            return MoviePyBackend.is_available()
        except (ImportError, AttributeError, TypeError):
            logger.debug("MoviePy backend not available")
            return False
    elif backend_enum == Backend.PYAV:
        try:
            from .pyav import PyAVBackend
            return PyAVBackend.is_available()
        except (ImportError, AttributeError, TypeError):
            logger.debug("PyAV backend not available")
            return False
    elif backend_enum == Backend.GSTREAMER:
        try:
            from .gstreamer import GStreamerBackend
            return GStreamerBackend.is_available()
        except (ImportError, AttributeError, TypeError):
            logger.debug("GStreamer backend not available")
            return False
    elif backend_enum == Backend.VLC:
        try:
            from .vlc import VLCBackend
            return VLCBackend.is_available()
        except (ImportError, AttributeError, TypeError):
            logger.debug("VLC backend not available")
            return False
    elif backend_enum == Backend.DUMMY:
        return True
    else:
        return False



"""
Screen capture functionality for FMUS-VID.

This module provides tools for capturing screenshots of the screen or specific regions.
"""

import os
import platform
from pathlib import Path
from typing import Optional, Tuple, Union, List
import tempfile
import time

import numpy as np
from PIL import Image, ImageGrab
import cv2  # Added for image processing

# Define a simple Frame class for the capture module instead of importing from core
class Frame:
    """Simplified Frame class for captured screenshots."""

    def __init__(self, frame_data):
        """Initialize a Frame with numpy array data."""
        self.data = frame_data

    def save(self, path):
        """Save the frame to a file."""
        path = Path(path)
        cv2.imwrite(str(path), cv2.cvtColor(self.data, cv2.COLOR_RGB2BGR))
        return path

    def to_numpy(self):
        """Return the frame as a numpy array."""
        return self.data

    def to_pil(self):
        """Convert to PIL Image."""
        return Image.fromarray(self.data)


def capture_screen(output_path: Optional[Union[str, Path]] = None) -> Union[Frame, Path]:
    """
    Capture a screenshot of the entire screen.

    Args:
        output_path: Optional path to save the screenshot. If None, returns a Frame object.

    Returns:
        Frame object if output_path is None, otherwise the Path to the saved image.

    Example:
        >>> # Capture and get Frame object
        >>> frame = fmusvid.capture.capture_screen()
        >>> # Capture and save to file
        >>> path = fmusvid.capture.capture_screen("screenshot.png")
    """
    # Capture the screen
    screenshot = ImageGrab.grab()

    # Convert to numpy array for processing
    frame_array = np.array(screenshot)

    # Create a Frame object
    frame = Frame(frame_array)

    # If output path is provided, save the image
    if output_path:
        output_path = Path(output_path)
        return frame.save(output_path)

    return frame


def capture_region(region: Tuple[int, int, int, int],
                 output_path: Optional[Union[str, Path]] = None) -> Union[Frame, Path]:
    """
    Capture a screenshot of a specific region of the screen.

    Args:
        region: Tuple of (left, top, right, bottom) coordinates
        output_path: Optional path to save the screenshot. If None, returns a Frame object.

    Returns:
        Frame object if output_path is None, otherwise the Path to the saved image.

    Example:
        >>> # Capture region (100, 100, 500, 400) and get Frame object
        >>> frame = fmusvid.capture.capture_region((100, 100, 500, 400))
        >>> # Capture region and save to file
        >>> path = fmusvid.capture.capture_region((100, 100, 500, 400), "region.png")
    """
    # Validate region coordinates
    left, top, right, bottom = region
    if left >= right or top >= bottom:
        raise ValueError("Invalid region coordinates. Ensure left < right and top < bottom.")

    # Capture the region
    screenshot = ImageGrab.grab(bbox=region)

    # Convert to numpy array for processing
    frame_array = np.array(screenshot)

    # Create a Frame object
    frame = Frame(frame_array)

    # If output path is provided, save the image
    if output_path:
        output_path = Path(output_path)
        return frame.save(output_path)

    return frame


def select_region() -> Tuple[int, int, int, int]:
    """
    Interactive tool to select a region of the screen.

    Returns:
        Tuple of (left, top, right, bottom) coordinates

    Example:
        >>> # Select a region interactively
        >>> region = fmusvid.capture.select_region()
        >>> # Capture the selected region
        >>> frame = fmusvid.capture.capture_region(region)
    """
    try:
        import tkinter as tk
        from tkinter import simpledialog
    except ImportError:
        raise ImportError("Tkinter is required for interactive region selection. "
                         "Install it or specify region coordinates manually.")

    # Create a full screen transparent window
    root = tk.Tk()
    root.attributes('-alpha', 0.3)
    root.attributes('-fullscreen', True)

    # Variables to store the selection coordinates
    start_x = start_y = end_x = end_y = 0
    selection_made = False

    # Create a canvas
    canvas = tk.Canvas(root, cursor="cross")
    canvas.pack(fill=tk.BOTH, expand=True)

    # Rectangle ID for the selection
    rect_id = None

    def on_mouse_down(event):
        nonlocal start_x, start_y, rect_id
        start_x, start_y = event.x, event.y
        # Create a rectangle if it doesn't exist
        if rect_id:
            canvas.delete(rect_id)
        rect_id = canvas.create_rectangle(start_x, start_y, start_x, start_y,
                                         outline='red', width=2)

    def on_mouse_move(event):
        nonlocal rect_id
        if rect_id:
            canvas.coords(rect_id, start_x, start_y, event.x, event.y)

    def on_mouse_up(event):
        nonlocal end_x, end_y, selection_made
        end_x, end_y = event.x, event.y
        selection_made = True
        root.quit()

    # Bind mouse events
    canvas.bind("<ButtonPress-1>", on_mouse_down)
    canvas.bind("<B1-Motion>", on_mouse_move)
    canvas.bind("<ButtonRelease-1>", on_mouse_up)

    # Run the main loop
    root.mainloop()

    # Clean up
    root.destroy()

    # Sort coordinates so left < right and top < bottom
    left = min(start_x, end_x)
    top = min(start_y, end_y)
    right = max(start_x, end_x)
    bottom = max(start_y, end_y)

    return (left, top, right, bottom)



#!/usr/bin/env python
"""
RTMP Demo Script

This script demonstrates the RTMP streaming capabilities of FMUS-VID
using a local video file as a source (as a stand-in for a live RTMP stream).

It simulates the streaming process and shows how real-time transcription,
translation, and subtitle overlay would work in a production environment.
"""

import os
import sys
import asyncio
import time
import logging
from pathlib import Path
import numpy as np
import cv2

# Add parent directory to path for imports
sys.path.insert(0, str(Path(__file__).resolve().parent.parent))

# Setup logging
logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s - %(name)s - %(levelname)s - %(message)s'
)
logger = logging.getLogger(__name__)

# Check if required modules are available
try:
    from fmusvid.stream import RTMPCapture, RTMPOutput
    from fmusvid.ai.speech import WhisperTranscriber
    from fmusvid.ai.translation import LLMTranslator
    from fmusvid.operations.text import SubtitleOverlay

    MODULES_AVAILABLE = True
except ImportError as e:
    logger.warning(f"RTMP modules not fully available: {e}")
    MODULES_AVAILABLE = False

# Default values
DEFAULT_VIDEO_PATH = os.path.join("examples", "assets", "sample.mp4")
DEFAULT_OUTPUT_DIR = os.path.join("examples", "output")
DEFAULT_LANGUAGES = ["en", "es", "fr", "de"]
DEFAULT_FPS = 30.0

class VideoFileSimulator:
    """
    Simulates an RTMP stream using a local video file.
    This is a stand-in for the RTMPCapture class when no actual RTMP server is available.
    """
    def __init__(self, video_path, fps=30.0):
        self.video_path = video_path
        self.target_fps = fps
        self.cap = None
        self.is_running = False
        self.frame_count = 0
        self.audio_data = np.zeros((1024, 1), dtype=np.float32)  # Dummy audio

    async def start(self):
        """Start the video capture."""
        self.cap = cv2.VideoCapture(self.video_path)
        if not self.cap.isOpened():
            raise ValueError(f"Could not open video file: {self.video_path}")
        self.is_running = True
        self.frame_count = 0
        logger.info(f"Started video file simulator: {self.video_path}")
        return True

    async def stop(self):
        """Stop the video capture."""
        if self.cap:
            self.cap.release()
        self.is_running = False
        logger.info("Stopped video file simulator")

    async def get_frame(self):
        """Get the next frame from the video."""
        if not self.is_running or not self.cap:
            return None

        ret, frame = self.cap.read()
        if not ret:
            return None

        self.frame_count += 1
        return frame

    async def get_audio(self):
        """Simulate getting audio from the video."""
        # In a real implementation, we would extract audio from the video
        # Here we just return dummy audio data
        return self.audio_data

    async def aiter_frames_with_audio(self):
        """Async iterator for frames and audio."""
        while self.is_running:
            frame = await self.get_frame()
            if frame is None:
                break

            audio = await self.get_audio()

            # Simulate real-time playback
            await asyncio.sleep(1/self.target_fps)

            yield frame, audio


class VideoFileOutput:
    """
    Simulates an RTMP output using a local video file.
    This is a stand-in for the RTMPOutput class when no actual RTMP server is available.
    """
    def __init__(self, output_path, width=1280, height=720, fps=30.0):
        self.output_path = output_path
        self.width = width
        self.height = height
        self.fps = fps
        self.writer = None
        self.is_running = False

    async def start(self):
        """Start the video writer."""
        # Ensure output directory exists
        os.makedirs(os.path.dirname(self.output_path), exist_ok=True)

        fourcc = cv2.VideoWriter_fourcc(*'mp4v')
        self.writer = cv2.VideoWriter(
            self.output_path,
            fourcc,
            self.fps,
            (self.width, self.height)
        )
        self.is_running = True
        logger.info(f"Started video file output: {self.output_path}")
        return True

    async def stop(self):
        """Stop the video writer."""
        if self.writer:
            self.writer.release()
        self.is_running = False
        logger.info(f"Stopped video file output: {self.output_path}")

    async def write_frame(self, frame):
        """Write a frame to the output video."""
        if not self.is_running or not self.writer:
            return False

        # Resize frame if needed
        if frame.shape[1] != self.width or frame.shape[0] != self.height:
            frame = cv2.resize(frame, (self.width, self.height))

        self.writer.write(frame)
        return True


class DummyTranscriber:
    """
    Dummy transcriber class for demonstration purposes.
    """
    def __init__(self):
        self.counter = 0
        self.transcripts = [
            "Hello, welcome to the RTMP streaming demo.",
            "This demonstrates real-time transcription and translation.",
            "We're simulating audio transcription right now.",
            "In a real implementation, we would use Whisper for this.",
            "The transcription would be more accurate and responsive.",
            "This is just a placeholder for demonstration purposes."
        ]

    async def transcribe_chunk(self, audio):
        """Simulate transcribing audio chunks."""
        # Only return text every 30 frames (about 1 second at 30fps)
        if self.counter % 30 == 0:
            text = self.transcripts[(self.counter // 30) % len(self.transcripts)]
            self.counter += 1
            return text
        self.counter += 1
        return ""


class DummyTranslator:
    """
    Dummy translator class for demonstration purposes.
    """
    def __init__(self, target_languages=None):
        self.target_languages = target_languages or ["es", "fr", "de"]
        self.translations = {
            "es": {
                "Hello, welcome to the RTMP streaming demo.": "Hola, bienvenido a la demostración de streaming RTMP.",
                "This demonstrates real-time transcription and translation.": "Esto demuestra la transcripción y traducción en tiempo real.",
                "We're simulating audio transcription right now.": "Estamos simulando la transcripción de audio ahora mismo.",
                "In a real implementation, we would use Whisper for this.": "En una implementación real, usaríamos Whisper para esto.",
                "The transcription would be more accurate and responsive.": "La transcripción sería más precisa y receptiva.",
                "This is just a placeholder for demonstration purposes.": "Esto es solo un marcador de posición con fines de demostración."
            },
            "fr": {
                "Hello, welcome to the RTMP streaming demo.": "Bonjour, bienvenue à la démo de streaming RTMP.",
                "This demonstrates real-time transcription and translation.": "Ceci démontre la transcription et la traduction en temps réel.",
                "We're simulating audio transcription right now.": "Nous simulons la transcription audio en ce moment.",
                "In a real implementation, we would use Whisper for this.": "Dans une implémentation réelle, nous utiliserions Whisper pour cela.",
                "The transcription would be more accurate and responsive.": "La transcription serait plus précise et réactive.",
                "This is just a placeholder for demonstration purposes.": "Ce n'est qu'un espace réservé à des fins de démonstration."
            },
            "de": {
                "Hello, welcome to the RTMP streaming demo.": "Hallo, willkommen zur RTMP-Streaming-Demo.",
                "This demonstrates real-time transcription and translation.": "Dies demonstriert Echtzeit-Transkription und -Übersetzung.",
                "We're simulating audio transcription right now.": "Wir simulieren gerade die Audiotranskription.",
                "In a real implementation, we would use Whisper for this.": "In einer realen Implementierung würden wir dafür Whisper verwenden.",
                "The transcription would be more accurate and responsive.": "Die Transkription wäre genauer und reaktionsschneller.",
                "This is just a placeholder for demonstration purposes.": "Dies ist nur ein Platzhalter für Demonstrationszwecke."
            }
        }

    async def translate_async(self, text):
        """Simulate translating text."""
        result = {}
        for lang in self.target_languages:
            if lang in self.translations and text in self.translations[lang]:
                result[lang] = self.translations[lang][text]
            else:
                # Fallback translation
                result[lang] = f"[{lang}] {text}"
        return result


class SimpleSubtitleOverlay:
    """
    Simple subtitle overlay class for demonstration purposes.
    """
    def __init__(self, position="bottom", font_size=28, background_opacity=0.7):
        self.position = position
        self.font_size = font_size
        self.background_opacity = background_opacity

    def apply(self, frame, text):
        """Apply subtitle overlay to frame."""
        if not text:
            return frame.copy()

        # Copy frame to avoid modifying original
        result = frame.copy()

        # Get frame dimensions
        height, width = result.shape[:2]

        # Calculate text size
        font = cv2.FONT_HERSHEY_SIMPLEX
        font_scale = self.font_size / 30.0
        thickness = 1
        (text_width, text_height), baseline = cv2.getTextSize(
            text, font, font_scale, thickness
        )

        # Determine text position
        if self.position == "bottom":
            text_x = (width - text_width) // 2
            text_y = height - 30
        else:
            text_x = (width - text_width) // 2
            text_y = 50

        # Draw background
        pad = 10
        bg_rect = (
            text_x - pad,
            text_y - text_height - pad,
            text_width + 2 * pad,
            text_height + 2 * pad
        )

        overlay = result.copy()
        cv2.rectangle(
            overlay,
            (bg_rect[0], bg_rect[1]),
            (bg_rect[0] + bg_rect[2], bg_rect[1] + bg_rect[3]),
            (0, 0, 0),
            -1
        )

        # Apply overlay with transparency
        cv2.addWeighted(
            overlay, self.background_opacity,
            result, 1 - self.background_opacity,
            0, result
        )

        # Draw text
        cv2.putText(
            result,
            text,
            (text_x, text_y),
            font,
            font_scale,
            (255, 255, 255),
            thickness
        )

        return result


async def process_video_file(
    video_path,
    output_dir,
    languages=None,
    fps=30.0,
    max_duration=60.0  # 1 minute max by default
):
    """
    Process a video file and generate outputs with subtitles in multiple languages.

    Args:
        video_path: Path to the input video file
        output_dir: Directory to save output videos
        languages: List of language codes to generate subtitles for
        fps: Output frame rate
        max_duration: Maximum duration to process in seconds
    """
    languages = languages or DEFAULT_LANGUAGES
    os.makedirs(output_dir, exist_ok=True)

    logger.info(f"Processing video file: {video_path}")
    logger.info(f"Output directory: {output_dir}")
    logger.info(f"Languages: {', '.join(languages)}")

    try:
        # Create video file simulator (instead of RTMP capture)
        video_simulator = VideoFileSimulator(video_path, fps=fps)
        await video_simulator.start()

        # Get video dimensions
        cap = cv2.VideoCapture(video_path)
        width = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
        height = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
        cap.release()

        # Create dummy transcriber and translator
        transcriber = DummyTranscriber()
        translator = DummyTranslator(target_languages=[lang for lang in languages if lang != "en"])

        # Create subtitle overlays
        subtitle_overlays = {
            lang: SimpleSubtitleOverlay(
                position="bottom",
                font_size=28,
                background_opacity=0.7
            )
            for lang in languages
        }

        # Create video outputs
        video_outputs = {}
        for lang in languages:
            output_path = os.path.join(output_dir, f"output_{lang}.mp4")
            video_outputs[lang] = VideoFileOutput(
                output_path=output_path,
                width=width,
                height=height,
                fps=fps
            )
            await video_outputs[lang].start()

        # Process frames
        start_time = time.time()
        frame_count = 0
        current_text = {"en": ""}
        subtitle_display_time = 0

        # Main processing loop
        async for frame, audio in video_simulator.aiter_frames_with_audio():
            try:
                # Check if we've reached the maximum duration
                elapsed = time.time() - start_time
                if elapsed > max_duration:
                    logger.info(f"Reached maximum duration of {max_duration}s")
                    break

                # Process audio for transcription
                text = await transcriber.transcribe_chunk(audio)
                if text:
                    logger.info(f"Transcribed: {text}")
                    current_text["en"] = text
                    subtitle_display_time = time.time() + 3.0  # Display for 3 seconds

                    # Translate to other languages
                    if text.strip():
                        translations = await translator.translate_async(text)
                        # Update current text for each language
                        for lang, translated_text in translations.items():
                            if lang in languages and lang != "en":
                                current_text[lang] = translated_text
                                logger.info(f"Translated ({lang}): {translated_text}")

                # Process frames for each language
                for lang in languages:
                    # Get text for this language
                    text = current_text.get(lang, "")

                    # Clear subtitle if display time has elapsed
                    if time.time() > subtitle_display_time:
                        text = ""

                    # Apply subtitle overlay
                    if text and lang in subtitle_overlays:
                        frame_with_subtitle = subtitle_overlays[lang].apply(frame, text)
                    else:
                        frame_with_subtitle = frame.copy()

                    # Write frame to output
                    if lang in video_outputs:
                        await video_outputs[lang].write_frame(frame_with_subtitle)

                # Update stats
                frame_count += 1
                if frame_count % 100 == 0:
                    fps_actual = frame_count / (time.time() - start_time)
                    logger.info(f"Processed {frame_count} frames ({fps_actual:.2f} fps)")

            except Exception as e:
                logger.error(f"Error processing frame: {e}")

        # Clean up
        for output in video_outputs.values():
            await output.stop()

        await video_simulator.stop()

        logger.info(f"Processing complete. Processed {frame_count} frames in {time.time() - start_time:.2f}s")
        logger.info(f"Output videos saved to {output_dir}")

        # Print output video paths
        for lang in languages:
            output_path = os.path.join(output_dir, f"output_{lang}.mp4")
            logger.info(f"{lang} output: {output_path}")

    except Exception as e:
        logger.error(f"Error in video processing: {e}")


def main():
    """Main function to run the demo."""
    import argparse

    parser = argparse.ArgumentParser(description="RTMP Demo using local video files")
    parser.add_argument("--video", default=DEFAULT_VIDEO_PATH, help="Path to input video file")
    parser.add_argument("--output", default=DEFAULT_OUTPUT_DIR, help="Output directory")
    parser.add_argument("--languages", default=",".join(DEFAULT_LANGUAGES), help="Comma-separated list of language codes")
    parser.add_argument("--fps", type=float, default=DEFAULT_FPS, help="Output frame rate")
    parser.add_argument("--duration", type=float, default=60.0, help="Maximum duration to process in seconds")

    args = parser.parse_args()

    print("\nRTMP Demo Script\n")
    print("This script demonstrates the RTMP streaming capabilities of FMUS-VID")
    print("using a local video file as a source.")
    print("\nFeatures demonstrated:")
    print("- Video stream processing")
    print("- Simulated real-time audio transcription")
    print("- Simulated translation to multiple languages")
    print("- Subtitle overlay")
    print("- Multi-language video output")

    # Create output directory
    os.makedirs(args.output, exist_ok=True)

    # Check if video file exists
    if not os.path.exists(args.video):
        print(f"\nError: Video file not found: {args.video}")
        print("You can specify a different video file with --video")
        return

    try:
        # Run the async processing function
        asyncio.run(process_video_file(
            video_path=args.video,
            output_dir=args.output,
            languages=args.languages.split(","),
            fps=args.fps,
            max_duration=args.duration
        ))

        print("\nDemo completed successfully!")
        print(f"Output videos saved to: {args.output}")

    except KeyboardInterrupt:
        print("\nDemo stopped by user")
    except Exception as e:
        print(f"\nError running demo: {e}")


if __name__ == "__main__":
    main()



"""
Performance optimization utilities for FMUS-VID.

This module provides tools for improving processing speed and memory efficiency.
"""

import os
import logging
from typing import Dict, List, Optional, Tuple, Callable, Any
import numpy as np
from pathlib import Path
import threading
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor
import torch
import cupy
from dataclasses import dataclass
from collections import OrderedDict

logger = logging.getLogger(__name__)

@dataclass
class BatchConfig:
    """Configuration for batch processing."""
    batch_size: int = 32
    num_workers: int = os.cpu_count() or 1
    use_multiprocessing: bool = False
    prefetch_factor: int = 2

class BatchProcessor:
    """
    Handles batch processing of video frames.
    """

    def __init__(self, config: Optional[BatchConfig] = None):
        """
        Initialize batch processor.

        Args:
            config: Batch processing configuration
        """
        self.config = config or BatchConfig()
        self._executor = (
            ProcessPoolExecutor(max_workers=self.config.num_workers)
            if self.config.use_multiprocessing
            else ThreadPoolExecutor(max_workers=self.config.num_workers)
        )

    def process_frames(
        self,
        frames: List[np.ndarray],
        operation: Callable[[np.ndarray], np.ndarray]
    ) -> List[np.ndarray]:
        """
        Process frames in batches.

        Args:
            frames: List of input frames
            operation: Function to apply to each frame

        Returns:
            List of processed frames
        """
        results = []
        for i in range(0, len(frames), self.config.batch_size):
            batch = frames[i:i + self.config.batch_size]
            futures = [
                self._executor.submit(operation, frame)
                for frame in batch
            ]
            results.extend([future.result() for future in futures])
        return results

    def __del__(self):
        """Cleanup executor."""
        self._executor.shutdown()

class FrameCache:
    """
    Cache for frequently accessed frames.
    """

    def __init__(self, max_size_mb: int = 1024):
        """
        Initialize frame cache.

        Args:
            max_size_mb: Maximum cache size in megabytes
        """
        self.max_size = max_size_mb * 1024 * 1024  # Convert to bytes
        self._cache: OrderedDict[str, np.ndarray] = OrderedDict()
        self._size = 0
        self._lock = threading.Lock()

    def get(self, key: str) -> Optional[np.ndarray]:
        """
        Get frame from cache.

        Args:
            key: Cache key

        Returns:
            Cached frame if found, None otherwise
        """
        with self._lock:
            if key in self._cache:
                # Memindahkan item yang diakses ke akhir (MRU)
                frame = self._cache.pop(key)
                self._cache[key] = frame
                return frame
        return None

    def put(self, key: str, frame: np.ndarray) -> None:
        """
        Add frame to cache.

        Args:
            key: Cache key
            frame: Frame to cache
        """
        frame_size = frame.nbytes

        with self._lock:
            # Menghapus frame lama jika cache penuh
            while self._size + frame_size > self.max_size and self._cache:
                old_key, old_frame = self._cache.popitem(last=False)
                self._size -= old_frame.nbytes

            # Menambahkan frame baru
            self._cache[key] = frame
            self._size += frame_size

    def clear(self) -> None:
        """Clear cache."""
        with self._lock:
            self._cache.clear()
            self._size = 0

class LazyFrameLoader:
    """
    Lazy loading of video frames with preloading support.
    """

    def __init__(
        self,
        video_path: Path,
        preload_seconds: float = 2.0,
        max_preload_frames: int = 60
    ):
        """
        Initialize lazy frame loader.

        Args:
            video_path: Path to video file
            preload_seconds: Seconds of video to preload
            max_preload_frames: Maximum number of frames to preload
        """
        self.video_path = Path(video_path)
        self.preload_seconds = preload_seconds
        self.max_preload_frames = max_preload_frames
        self._preloaded: Dict[int, np.ndarray] = {}
        self._current_frame = 0
        self._preload_thread: Optional[threading.Thread] = None
        self._lock = threading.Lock()

    def get_frame(self, frame_number: int) -> np.ndarray:
        """
        Get frame by number, using preloaded frame if available.

        Args:
            frame_number: Frame number to retrieve

        Returns:
            Video frame
        """
        with self._lock:
            if frame_number in self._preloaded:
                frame = self._preloaded.pop(frame_number)
                self._current_frame = frame_number
                self._start_preload()
                return frame

        # Memuat frame jika tidak ada di cache
        frame = self._load_frame(frame_number)
        self._current_frame = frame_number
        self._start_preload()
        return frame

    def _start_preload(self) -> None:
        """Start preloading thread."""
        if (
            self._preload_thread is None
            or not self._preload_thread.is_alive()
        ):
            self._preload_thread = threading.Thread(
                target=self._preload_frames
            )
            self._preload_thread.daemon = True
            self._preload_thread.start()

    def _preload_frames(self) -> None:
        """Preload upcoming frames."""
        start = self._current_frame + 1
        for i in range(start, start + self.max_preload_frames):
            if len(self._preloaded) >= self.max_preload_frames:
                break
            try:
                frame = self._load_frame(i)
                with self._lock:
                    self._preloaded[i] = frame
            except Exception as e:
                logger.debug(f"Preload stopped at frame {i}: {e}")
                break

    def _load_frame(self, frame_number: int) -> np.ndarray:
        """
        Load single frame from video.

        Args:
            frame_number: Frame number to load

        Returns:
            Video frame
        """
        # Implementasi sebenarnya akan menggunakan backend video
        raise NotImplementedError

class GPUAccelerator:
    """
    GPU acceleration utilities.
    """

    def __init__(self):
        """Initialize GPU accelerator."""
        self.cuda_available = torch.cuda.is_available()
        self.cuda_device_count = (
            torch.cuda.device_count() if self.cuda_available else 0
        )

        try:
            import cupy as cp
            self.cupy_available = True
        except ImportError:
            self.cupy_available = False

    def get_optimal_device(self) -> str:
        """
        Get optimal compute device.

        Returns:
            Device name ('cuda', 'cpu')
        """
        if self.cuda_available and self.cuda_device_count > 0:
            return 'cuda'
        return 'cpu'

    def get_device_info(self) -> Dict[str, Any]:
        """
        Get detailed device information.

        Returns:
            Dictionary containing device information
        """
        info = {
            'cuda_available': self.cuda_available,
            'device_count': self.cuda_device_count,
            'cupy_available': self.cupy_available,
        }

        if self.cuda_available:
            info.update({
                'current_device': torch.cuda.current_device(),
                'device_name': torch.cuda.get_device_name(),
                'memory_allocated': torch.cuda.memory_allocated(),
                'memory_cached': torch.cuda.memory_reserved(),
            })

        return info

    def to_gpu(self, data: np.ndarray) -> Any:
        """
        Transfer data to GPU memory.

        Args:
            data: Input data

        Returns:
            Data in GPU memory
        """
        if self.cuda_available:
            if self.cupy_available:
                return cupy.asarray(data)
            return torch.from_numpy(data).cuda()
        return data

    def to_cpu(self, data: Any) -> np.ndarray:
        """
        Transfer data to CPU memory.

        Args:
            data: Input data

        Returns:
            Data in CPU memory
        """
        if isinstance(data, cupy.ndarray):
            return cupy.asnumpy(data)
        if torch.is_tensor(data):
            return data.cpu().numpy()
        return data

class MemoryManager:
    """
    Memory usage optimization.
    """

    def __init__(self, target_usage_mb: int = 1024):
        """
        Initialize memory manager.

        Args:
            target_usage_mb: Target memory usage in megabytes
        """
        self.target_usage = target_usage_mb * 1024 * 1024  # Convert to bytes
        self._cached_objects: Dict[str, Any] = {}

    def register_object(self, key: str, obj: Any) -> None:
        """
        Register object for memory management.

        Args:
            key: Object identifier
            obj: Object to manage
        """
        self._cached_objects[key] = obj

    def clear_cache(self) -> None:
        """Clear all cached objects."""
        self._cached_objects.clear()

        if torch.cuda.is_available():
            torch.cuda.empty_cache()

        if cupy.cuda.runtime.getDeviceCount() > 0:
            mempool = cupy.get_default_memory_pool()
            mempool.free_all_blocks()
